#include <glut.h>         
#include <math.h>
#include <iostream>
#include <stdlib.h>
#include <string>
#include <stdio.h>
#include <vector>

#include "Object.h"
#include "Mobile.h"
#include "HeightMap.h"
#include "mouse.h"
#include "glm.h"
#include "ObjectList.h"
#include "Audio.h"


using namespace std;

int mX = 0;
int mY = 0;
int mZ = 0;
int prevButton1State = 0;
int prevButton2State = 0;

// stuff for lighting
GLfloat lAmbient[] = {0.3,0.3,0.3,1.0};
GLfloat lDiffuse[] = {1.0,1.0,1.0,1.0};
GLfloat lSpecular[] = {1.0,1.0,1.0,1.0};

GLfloat lPosition[] = {0.0,100.0,0.0,1.0};

// materials
GLfloat mSpecular[] = {1.0,1.0,1.0,1.0};
GLfloat mShininess[] = {128.0};

//colors
GLfloat cBlack[] = {0.0,0.0,0.0,1.0};
GLfloat cOrange[] = {1.0,0.5,0.5,1.0}; 
GLfloat cWhite[] = {1.0,1.0,1.0,1.0}; 
GLfloat cGrey[] = {0.1,0.1,0.1,1.0};
GLfloat cLightGrey[] = {0.9,0.9,0.9,1.0};


// window aspect ratio
float aspectRatio;
// window size
int h, w;

// current position of mobile object
float my_x=0.0f,my_y=0.0f,my_z=0.0f;

// mouselook eyeVector
triple eyeVector;

// for mouselook
float xAngle, yAngle, xAngle_temp, yAngle_temp;

// objects
HeightMap map;
Mouse mouse;
ObjectList objects;

// laser
triple source;
triple target;
bool laser = false;
bool gameover = false;

// ground texture
GLuint ground = 0;

// sound stuff here
unsigned int MAX_SOUNDS = 3;
Audio audio;
Sound* soundList;
unsigned int SCounter = 0;

void initializeSounds(){
	soundList = new Sound[MAX_SOUNDS];
	// first sound is the background midi
	soundList[0] = Sound("data/halo_theme_v2.mid");
	soundList[1] = Sound("data/rocket_launcher.wav");
	soundList[2] = Sound("data/laser1.wav");
	
	/*
	Removed for simplicity's sake
	// now, a gun sound
	soundList[1] = Sound("data/gun1.wav");
	// another gun sound
	soundList[2] = Sound("data/gun2.wav");
	// yet another gun sound
	soundList[3] = Sound("data/gun3.wav");
	// explosion sound
	soundList[4] = Sound("data/explosion1.wav");
	// another explosion sound
	soundList[5] = Sound("data/explosion.wav");
	// a rocket launching sound
	soundList[6] = Sound("data/rocket.wav");
	*/ 
}

void* loadPPM( const char *filename, int& width, int &height, int &numComponents );

// loadTexture implementation.  You will need to modify this function if you use mipmaps.
unsigned int loadTexture( const char* fileName ) {
	int w,h,nComp;
	void* pImage = loadPPM(fileName,w,h,nComp);
	if (pImage == NULL ) {
		printf( "error: Image %s not loaded\n", fileName );
		return 0;
	}
	// setup textures...
	glEnable( GL_TEXTURE_2D );
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	unsigned int nTexID;
	glGenTextures( 1, &nTexID );
    glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	glBindTexture(GL_TEXTURE_2D, nTexID);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	if ( nComp == 1 ) {
		glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE8, w, h, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, pImage);  
	} else if ( nComp == 3 ) {
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGB, GL_UNSIGNED_BYTE, pImage);
	} else if ( nComp == 4 ) {
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, pImage);
	} else {
		printf("error: I don't know what to do with this image format\n");
	}
	glDisable( GL_TEXTURE_2D );
	delete pImage;
	return nTexID;
}

// Read image file.. (ppm)
void* loadPPM( const char *filename, int& width, int &height, int &numComponents ) {
	FILE          *imgFile;
	char          buf[1024];
	char          type;
	int           i, j, k, l;
	GLubyte*      imageData;

	imgFile= fopen( filename, "rb" );
	if( errno != 0 )
		//return true;
		return NULL;
	// read PNM magic number (P1 to P8, where P7 and P8 are non-standard)
	if( fscanf( imgFile, "P%c \n", &type )!= 1 || type< '1' || type> '8')
		//  return true;
		return NULL;
	while( fscanf( imgFile, "#%[^\n]\n", buf ) )  // skip comments
	;
	fscanf( imgFile, "%d", &width );  // read width
	while( fscanf( imgFile, "#%[^\n]\n", buf ) )  // skip comments
	;
	fscanf( imgFile, "%d", &height ); // read height
	while( fscanf( imgFile, "#%[^\n]\n", buf ) ) // skip comments
	;
	// skip max. component and exactly one whitespace
	fscanf( imgFile, "%*d%*c" );

	switch( type ){
	case '1': // ASCII bitmap
	case '4': // binary bitmap
		std::cerr << "Bitmaps not implemented\n";
		//return true;
		return NULL;
	case '2': // ASCII greymap
		imageData= new GLubyte[width*height*3];
		for( i= 0 ; i< height ; i++ ) {
			for( j= 0 ; j< width ; j++ ) {
				fscanf( imgFile, "%d", &l );
				imageData[i*width+j]= l;
			}
		}
		numComponents= 1;
		break;
	case '3': // ASCII RGB
		imageData= new GLubyte[width*height*3];
		for( i= 0 ; i< height ; i++ ) {
			for( j= 0 ; j< width ; j++ ) {
				for( k= 0 ; k< 3 ; k++ ) {
					fscanf( imgFile, "%d", &l );
					imageData[(i*width+j)*3+k]= l;
				}
			}
		}
		numComponents= 3;
		break;
	case '5': // binary greymap
		imageData= new GLubyte[width*height];
		fread( imageData, 1, width*height, imgFile );
		numComponents= 1;
		break;
	case '6': // binary RGB
		imageData= new GLubyte[width*height*3];
		fread( imageData, 1, width*height*3, imgFile );
		numComponents= 3;
		break;
	case '7': // ASCII RGBA: this is not a standard format!
		imageData= new GLubyte[width*height*4];
		for( i= 0 ; i< height ; i++ ) {
			for( j= 0 ; j< width ; j++ ){
				for( k= 0 ; k< 4 ; k++ ) {
					fscanf( imgFile, "%d", &l );
					imageData[(i*width+j)*4+k]= l;
				}
			}
		}
		numComponents= 4;
		break;
	case '8': // binary RGBA: this is not a standard format!
		imageData= new GLubyte[width*height*4];
		fread( imageData, 1, width*height*4, imgFile );
		numComponents= 4;
		break;
	}
	fclose( imgFile );
	return imageData;
}



void changeSize(int new_w, int new_h)
	{

	// avoid div by 0
	if(new_h == 0)
		new_h = 1;

	w = new_w;
	h = new_h;
	aspectRatio = 1.0f * w / h;
	// Reset the coordinate system before modifying
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	// Set the viewport to be the entire window
    glViewport(0, 0, w, h);

	// Set the clipping volume
	gluPerspective(45,aspectRatio,0.1,1000);

	// setting the camera now
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(my_x, my_y, my_z, 
		      my_x + eyeVector.x,my_y + eyeVector.y,my_z + eyeVector.z,
			  0.0f,1.0f,0.0f);
}


void initialize_graphics() {

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	
	glLightfv(GL_LIGHT0,GL_AMBIENT,lAmbient);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,lDiffuse);
	glLightfv(GL_LIGHT0,GL_SPECULAR,lSpecular);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_TRUE);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

    if (map.terrainLoadFromImage("data/map3.tga",1) != TERRAIN_OK) {
		std::cout << "ERROR LOADING MAP FILE, TERMINATING";
	  return;
	}
    map.terrainScale(0,80);
    map.loadMap();
	objects.loadFile("conf.txt");
	//ground = loadTexture("data/elephant.rot.ppm");
	ground = loadTexture("data/wood.ppm");

  eyeVector.x = 0.0f;
  eyeVector.y = 0.0f;
  eyeVector.z = -1.0f;



}


/* code for printing text to screen if necessary */
void setOrthographicProjection() {

	// switch to projection mode
	glMatrixMode(GL_PROJECTION);
	// save previous matrix which contains the 
	//settings for the perspective projection
	glPushMatrix();
	// reset matrix
	glLoadIdentity();
	// set a 2D orthographic projection
	glOrtho (0, w, h, 0, -99999, 99999);
	//gluOrtho2D(0, w, 0, h);
	// invert the y axis, down is positive
	glScalef(1, -1, 1);
	// mover the origin from the bottom left corner
	// to the upper left corner
	glTranslatef(0, -h, 0);
	glMatrixMode(GL_MODELVIEW);
}

/* code for printing text to screen if necessary */
void resetPerspectiveProjection() {
	// set the current matrix to GL_PROJECTION
	glMatrixMode(GL_PROJECTION);
	// restore previous settings
	glPopMatrix();
	// get back to GL_MODELVIEW matrix
	glMatrixMode(GL_MODELVIEW);
}

/* code for printing text to screen if necessary */
void renderBitmapString(float x, float y, void *font,char *string)
{
  
  char *c;
  // set position to start drawing fonts
  glRasterPos2f(x, y);
  // loop all the characters in the string
  for (c=string; *c != '\0'; c++) {
    glutBitmapCharacter(font, *c);
  }
}

// draw a translucent heads-up display
void drawHUD() {
	// HUD - need to write more code for HUD

	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	// activate blending
	glEnable (GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE);		
	setOrthographicProjection();
	glPushMatrix();
	glPushAttrib(GL_COLOR_BUFFER_BIT);
	glLoadIdentity();

	// code that actually draws the HUD

	glColor4f(0.3f,0.0f,1.0f,0.3f);	
	// draw crosshairs
	glRecti(w/2-10,h/2-10,w/2-7, h/2+10);	
	glRecti(w/2+7,h/2-10,w/2+10, h/2+10);
	
	// big bars
	glRecti(w/2-20,h/2-20,w/2-13, h/2+20);	
	glRecti(w/2+13,h/2-20,w/2+20, h/2+20);
	
	// draw health bar
	glColor4f(0.3f,0.0f,1.0f,0.3f);
	int worth = objects.playerWorth();
	int hits = objects.playerHitpoints();
	int redlength = (h*(worth-hits))/worth;
	glColor4f(1.0f,0.0f,0.0f,0.3f);
	glRecti(0,0,20,redlength);
	glColor4f(0.3f,0.0f,1.0f,0.3f);
	glRecti(0,redlength,20,h);

	// score
	glColor3f(1.0f,1.0f,1.0f);
	// draw game over if game over
	if (gameover) {
		renderBitmapString(w/2-65,h/2,GLUT_BITMAP_TIMES_ROMAN_24,"GAME OVER");
	}
	char score[100];
	sprintf(score,"Score: %d",objects.getScore());
	renderBitmapString(10,10,GLUT_BITMAP_8_BY_13,score);

	// restore everything
	glPopMatrix();
	glPopAttrib();
	resetPerspectiveProjection();
	glEnable(GL_LIGHTING);
	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
}

void drawLaser() {
	double p1[3] = {source.x,source.y,source.z};
	double p2[3] = {target.x,target.y,target.z};

	GLUquadricObj *cyl, *disk;  
	cyl  = gluNewQuadric();
	disk = gluNewQuadric();

	double pi = 3.1415926535;
	double r2d = 180.0/pi;	// radians to degrees conversion factor
	double sf = 1.5;		// scale factor
	double sfi = 1.0/sf;	// inverse scale factor

	double rb = 0.2;	// radius of cylinder bottom
	double rt = 0.2;	// radius of cylinder bottom
	int dth = 10;		// no. of angular cylinder subdivsions 
	int dz = 10;		// no. of cylinder subdivsions in z direction

	// orientation vector 
	double vx = p2[0]-p1[0];	//	component in x-direction
	double vy = p2[1]-p1[1];	//	component in y-direction
	double vz = p2[2]-p1[2];	//	component in z-direction

	double v = sqrt( vx*vx + vy*vy + vz*vz );	// cylinder length
		
	// rotation vector, z x r  
	double rx = -vy*vz;
	double ry = +vx*vz;
	double ax = 0.0;

	double zero = 1.0e-3;
	if (fabs(vz) < zero)
	{
		ax = r2d*acos( vx/v );	// rotation angle in x-y plane
		if ( vx <= 0.0 ) ax = -ax;
	}
	else
	{
		ax = r2d*acos( vz/v );	// rotation angle
		if ( vz <= 0.0 ) ax = -ax;
	}

	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	// activate blending
	glEnable (GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE);	
	glColor4f(1.0f, 0.0f, .1f,0.5f);	// set cylinder color
	
	glPushMatrix();

		glTranslated( p1[0], p1[1], p1[2] );	// translate to point 1

		if (fabs(vz) < zero)
		{
			glRotated(90.0, 0, 1, 0.0);			// Rotate & align with x axis
			glRotated(ax, -1.0, 0.0, 0.0);		// Rotate to point 2 in x-y plane
		}
		else
		{
			glRotated(ax, rx, ry, 0.0);			// Rotate about rotation vector
		}

		gluCylinder(cyl, rb, rt, v, dth, dz);	// draw side faces of cylinder
		gluDisk( disk, 0.0, rb, dth, 1);		// draw bottom of cylinder

	glPopMatrix();

	glPushMatrix();

		glTranslated( p2[0], p2[1], p2[2] );	// translate to point 2

		if (fabs(vz) < zero)
		{
			glRotated(90.0, 0, 1, 0.0);			// Rotate & align with x axis
			glRotated(ax, -1.0, 0.0, 0.0);		// Rotate to point 2 in the x-y plane
		}
		else
		{
			glRotated(ax, rx, ry, 0.0);			// Rotate about rotation vector
		}

		gluDisk( disk, 0.0, rt, dth, 1);		// draw top of cylinder
	
	glPopMatrix();
	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);

	gluDeleteQuadric( cyl );
	gluDeleteQuadric( disk );

	return;

}


void doDrawing(void) {

	glLoadIdentity();
	// set lookat point
	float x,y,z;
	objects.playerCoords(x,y,z);
	gluLookAt(x, y, z, 
		      x + 10*eyeVector.x,y + 10*eyeVector.y,z + 10*eyeVector.z,
			  0.0f,1.0f,0.0f);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLightfv(GL_LIGHT0,GL_POSITION,lPosition);


	// set material
	glMaterialfv(GL_FRONT, GL_SPECULAR, mSpecular);
	glMaterialfv(GL_FRONT, GL_SHININESS,mShininess);
	glMaterialfv(GL_FRONT, GL_AMBIENT, cGrey);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, cWhite);

	float planes[4];
	float planet[4];
	planes[0] = 1; planes[1] = 0; planes[2] = 0; planes[3] = 0;
	planet[0] = 0; planet[1] = 0; planet[2] = 1; planet[3] = 0;
	// draw call list for map
	glEnable (GL_TEXTURE_2D);
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_OBJECT_LINEAR);
	glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_OBJECT_LINEAR);

	//glColor3f(cLightGrey[0],cLightGrey[1],cLightGrey[2]);
	glTexGenfv(GL_S,GL_OBJECT_PLANE, planes);
	glTexGenfv(GL_T,GL_OBJECT_PLANE, planet);
	glBindTexture(GL_TEXTURE_2D, ground);
	glCallList(map.getCallStackNumber());
	glDisable(GL_TEXTURE_2D);

	// draw all objects
	objects.draw();


	if (laser) {
		drawLaser();
		soundList[2].playSFX();
	}

	drawHUD();
}

void doPicking();

void passiveMouse() {

	int button1 = 0;
	int button2 = 0;
	mouse.getMouseMovement(mX,mY,mZ,button1,button2);

	xAngle_temp = xAngle + (mX)*0.003;
	yAngle_temp = yAngle + (mY) * 0.003;
	if (yAngle_temp > 1.57)
		yAngle_temp = 1.57;
	else if (yAngle_temp < -1.57)
		yAngle_temp = -1.57;
	eyeVector.x = cos(yAngle_temp )*sin(xAngle_temp);
	eyeVector.z = -cos(yAngle_temp )*cos(xAngle_temp);
	eyeVector.y = -sin(yAngle_temp );

	xAngle = xAngle_temp;
	yAngle = yAngle_temp;

	if (button1 == 1 && prevButton1State == 0) {
		doPicking();
		prevButton1State = 1;
	}
	else {
		prevButton1State = button1;
	}

	if (button2 == 1  && prevButton2State == 0) {
		prevButton2State = 1;
		laser = true;
		doPicking();
	}
	else {
		laser = false;
		prevButton2State = button2;
	}

	if (laser) {
		float x,y,z;
		objects.playerCoords(x,y,z);
		laser = true;
	    target.x = x+300*eyeVector.x;
	    target.y = y+300*eyeVector.y;
	    target.z = z+300*eyeVector.z;

	    source.x = 5+x+10*eyeVector.x;
	    source.y = 5+y+10*eyeVector.y;
	    source.z = 5+z+10*eyeVector.z;
	}


	
}

void TimerCallback(int value) {

	if (objects.playerHitpoints() <= 0) {
		gameover = true;	
	}
	else {
		objects.move();
		objects.updateTimers();
		passiveMouse();	
	}
	glutTimerFunc(10,TimerCallback,1);
	glutPostRedisplay();
}

void KeyboardCallback(unsigned char key,int x, int y) {
	// ESC keypress
	if (key == 27) {
		exit(0);	
	}
	else if (key == 'q') {
		float x,y,z;
		objects.playerCoords(x,y,z);
		objects.addMissile(x,y,z,3);
	}
	else if (key == 'p') {
		// sound test...
		// loops through each sound in the SoundList
		soundList[SCounter].playBG();
		SCounter = (SCounter + 1) % MAX_SOUNDS;
	}

}

void draw_scene (void) {   // Function called every time scene is drawn.
  glClear(GL_COLOR_BUFFER_BIT);
  doDrawing();
  glutSwapBuffers ();  // Bring the back buffer to the front so we can see it.
}

// The processHits function is almost an exact copy of
// the red book sample picking function.
void processHits (GLint hits, GLuint buffer[])
{

	unsigned int i, j;
   GLuint names, *ptr;
   int name;
   float z1, oldZ;

   ptr = (GLuint *) buffer;
   for (i = 0; i < hits; i++) {  /* for each hit  */
      names = *ptr;
	  ptr++;
	  if (i != 0)
	  {
		  oldZ = z1;
	  }
	  z1 = (float) *ptr/0x7fffffff;
	  ptr++; 
	  ptr++;
	  if (i == 0)
	  {
		  name = *ptr;
	  }
	  else
      {
         if (z1 < oldZ)
		 {
			name = *ptr;
		 }
      }

      for (j = 0; j < names; j++) {  /* for each name */
			ptr++;
	  }
   }
	// look for the geometry that matched
   if ( (name > 0) && !laser) {
	  float x,y,z;
	  objects.playerCoords(x,y,z);
	  objects.addMissile(x,y,z,name);
	  soundList[1].playSFX();
   }
   else if ( (name > 0) && laser) {  
	  objects.hitEntity(name);

   }
}



// The doPicking function is almost an exact copy of
// the red book sample picking function.
void doPicking()
{
	int x = w/2;
	int y = h/2;
	GLuint selectBuf[256];
	GLint hits;
	GLint viewport[4];

	glGetIntegerv(GL_VIEWPORT, viewport);

	glSelectBuffer(256, selectBuf);
	glRenderMode(GL_SELECT);

	glInitNames();

	(void) glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	// picking region
	gluPickMatrix((GLdouble) x, (GLdouble) (viewport[3] - y),
                 5.0, 5.0, viewport);
	gluPerspective(60,aspectRatio,.1,1000); 
    glMatrixMode(GL_MODELVIEW);

	objects.draw();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glFlush();

	hits = glRenderMode(GL_RENDER);
	processHits(hits, selectBuf);
	
	glMatrixMode(GL_MODELVIEW) ;
} 


int main(int argc, char **argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
        

	glutInitWindowPosition(100,100);
	glutInitWindowSize(640,480);
	glutCreateWindow("Track DEMO");

	HWND hWnd = FindWindow("GLUT","Track DEMO");
	mouse.init(hWnd);
/*
	glutGameModeString("1024x768:32@70");
	if (glutGameModeGet(GLUT_GAME_MODE_POSSIBLE)) 
		glutEnterGameMode();
	else {
		printf("The select mode is not available\n");
		exit(1);	
	}
*/
	// register all callbacks
	glutDisplayFunc(draw_scene);
	glutTimerFunc(10,TimerCallback,1);
	glutReshapeFunc(changeSize);

	glutKeyboardFunc(KeyboardCallback);

	initialize_graphics();
	initializeSounds();
	// play background music (non-loopin, I think)
	soundList[0].playBG();


	glutMainLoop();

	return(0);
}
