#include <stdio.h>	
#include "Object.h"
#include "glm.h"
#include <string>
#include <math.h>
#include <iostream.h>


Node::Node() {
	for (int i = 0; i < 16; i++) {
		localMatrix[i] = 0;
		animateMatrix[i] = 0;
	}
	localMatrix[0] = localMatrix[5] = localMatrix[10] = localMatrix[15] = 1;
	animateMatrix[0] = animateMatrix[5] = animateMatrix[10] = animateMatrix[15] = 1;
	sibling = NULL;
	child = NULL;
	keyRotate[0] = keyRotate[1] = keyRotate[2] = 0;	
	callstackid = 0;
	keyframef.x = keyframef.y = keyframef.z = 0;
	keyframeb.x = keyframeb.y = keyframeb.z = 0;
	hitRadius = 1.0;
}

void
Node::loadMesh(char* fname) {
  GLMmodel* enemy = glmReadOBJ(fname);
  float scale = glmUnitize(enemy);
  glmFacetNormals(enemy);
  glmVertexNormals(enemy, 90.0);
  glmScale(enemy,1.0);
  callstackid = glmList(enemy,GLM_SMOOTH | GLM_MATERIAL);
  glmDelete(enemy);
}

void 
Object::readObjectConfig(Node *parent, ifstream &input ) {
	char token[512];
	while ( !input.eof() ) {
		input >> token;
		if ( token[0] == '#' ) {
			input.getline( token, 512, '\n' );
		} else if ( strcmp( token, "node" ) == 0 ) {
			input >> token; 
			Node *child = new Node();

			if (parent->child == NULL) {
				parent->child = child;
			}
			else {
				Node* currSibling = parent->child;
				while (currSibling->sibling != NULL) {
					currSibling = currSibling->sibling;
				}
				currSibling->sibling = child;
			}

			readObjectConfig( child, input );
		} else if ( strcmp( token, "mesh" ) == 0 ) {
			char filename[512];
			input >> filename;
			parent->loadMesh(filename);
		} else if ( strcmp( token, "translate" ) == 0 ) {
			double tx, ty, tz;
			input >> tx >> ty >> tz;
			glPushMatrix();
			glLoadIdentity();
			glMultMatrixf(parent->localMatrix);
			glTranslated(tx,ty,tz);
			glGetFloatv(GL_MODELVIEW_MATRIX, parent->localMatrix);
			glPopMatrix();
		} else if ( strcmp( token, "rotatex" ) == 0 ) {
			double theta;
			input >> theta;
			cout << theta;
			glPushMatrix();
			glLoadIdentity();
			glMultMatrixf(parent->localMatrix);
			glRotated(theta,1,0,0);
			glGetFloatv(GL_MODELVIEW_MATRIX, parent->localMatrix);
			glPopMatrix();
		} else if ( strcmp( token, "rotatey" ) == 0 ) {
			double theta;
			input >> theta;
			glPushMatrix();
			glLoadIdentity();
			glMultMatrixf(parent->localMatrix);
			glRotated(theta,0,1,0);
			glGetFloatv(GL_MODELVIEW_MATRIX, parent->localMatrix);
			glPopMatrix();
		} else if ( strcmp( token, "rotatez" ) == 0 ) {
			double theta;
			input >> theta;
			glPushMatrix();
			glLoadIdentity();
			glMultMatrixf(parent->localMatrix);
			glRotated(theta,0,0,1);
			glGetFloatv(GL_MODELVIEW_MATRIX, parent->localMatrix);
			glPopMatrix();
		} else if ( strcmp( token, "scale" ) == 0 ) {
			double s;
			input >> s;
			glPushMatrix();
			glLoadIdentity();
			glMultMatrixf(parent->localMatrix);
			glScaled(s,s,s);
			parent->hitRadius = s;
			glGetFloatv(GL_MODELVIEW_MATRIX, parent->localMatrix);
			glPopMatrix();
		} else if ( strcmp( token, "keyframef" ) == 0 ) {
			animated = true;
			float kx, ky, kz;
			input >> kx >> ky >> kz;
			parent->keyframef.x= kx;
			parent->keyframef.y = ky;
			parent->keyframef.z = kz;
		} else if ( strcmp( token, "keyframeb" ) == 0 ) {
			float kx, ky, kz;
			input >> kx >> ky >> kz;
			parent->keyframeb.x= kx;
			parent->keyframeb.y = ky;
			parent->keyframeb.z = kz;
		} else if ( strcmp( token, "}" ) == 0 ) {
			return;
		}
	}
}

void 
Object::readFile( char *fname ) {
	ifstream input(fname);
	if ( !input.is_open() ) {
		printf("couldn't open");
		return;
	}
	head = new Node();
	readObjectConfig( head, input );
	input.close();
}


Object::Object() {
	hittable = true;
	coordinates.x = coordinates.y = coordinates.z = 0;
	// initialize local and world transform matrices
	for (int i = 0; i < 16;i++) {
		if ((i % 5) != 0) {
			localTransform[i] = worldTransform[i] = 0;
		}
		else {
			localTransform[i] = worldTransform[i] = 1;				
		}
	}
	objectName = nameCount;
	nameCount++;
	animated = false;
	movable = false;
	hitpoints = 0;
	animateSpeed = 0;
	exploding = false;
	drawable = true;
	shootDelay = -1;
	shootTimer = 0;
	respawnDelay = -1;
	respawnTimer = 0;
	worth = 0;
}

Object::~Object() {
	destroy(head);
}

void 
Object::copyNodeHelper(Node* newNode, Node* oldNode) {
	if (oldNode == NULL) return;

	newNode->callstackid = oldNode->callstackid;
	newNode->hitRadius = oldNode->hitRadius;
	int i = 0;
	for (i = 0; i < 16; i++) {
		newNode->animateMatrix[i] = oldNode->animateMatrix[i];
		newNode->localMatrix[i] = oldNode->localMatrix[i];
	}

	if (oldNode->child != NULL) {
		newNode->child = new Node();
		copyNodeHelper(newNode->child, oldNode->child);
	}

	if (oldNode->sibling != NULL) {
		newNode->sibling = new Node();
		copyNodeHelper(newNode->sibling, oldNode->sibling);	

	}

}

Object::Object(Object& o) {
	hittable = o.hittable;
	coordinates.x = o.coordinates.x;
	coordinates.y = o.coordinates.y;
	coordinates.z = o.coordinates.z;
	animated = o.animated;
	movable = o.movable;
	animateSpeed = o.animateSpeed;
	hitpoints = o.hitpoints;
	head = new Node();

	copyNodeHelper(head,o.head);	
}

void
Object::copy(Object* o) {
	hittable = o->hittable;
	coordinates.x = o->coordinates.x;
	coordinates.y = o->coordinates.y;
	coordinates.z = o->coordinates.z;
	animated = o->animated;
	movable = o->movable;
	animateSpeed = o->animateSpeed;
	hitpoints = o->hitpoints;
	head = new Node();

	copyNodeHelper(head,o->head);
}

void 
Object::setLocalMatrix(float* m) {
	// copy matrix
	for (int i = 0; i < 16; i++ )
	{
		localTransform[i] = m[i];
	}
}

void 
Object::setWorldMatrix(float* m)
{
	// copy matrix
	for (int i = 0; i < 16; i++ )
	{
		worldTransform[i] = m[i];
	}
}

void 
Object::setHittable(bool hit) {
	hittable = hit;
}

void 
Object::setCallStackID(int id) {
	callstackNumber = id;
}

void 
Object::setCoordinates(float x, float y, float z) {
	coordinates.x = x;
	coordinates.y = y;
	coordinates.z = z;
}

void 
Object::setCoordinates(triple c) {
	coordinates.x = c.x;
	coordinates.y = c.y;
	coordinates.z = c.z;
}

float* 
Object::getLocalMatrix() {
	return localTransform;
}

float* 
Object::getWorldMatrix() {
	return worldTransform;
}

bool 
Object::isHittable() {
	return hittable;
}

int 
Object::getCallStackID() {
	return callstackNumber;
}

triple 
Object::getCoordinates() {
	return coordinates;
}

int 
Object::getObjectName() {
	return objectName;
}

void
Object::setObjectName(int name ) {
	objectName = name;
}

void 
Object::destroy(Node* n) {
	// traverse
	if (n == NULL) return;
	if (n->child != NULL) destroy(n->child);
	delete n->child;
	if (n->sibling != NULL) traverse(n->sibling);
	delete n->sibling;
}

void
Object::addHead(Node* n) {
	head = n;
}

void 
Object::animateDelta(Node* root,float inc) {
	// traverse
	if (root == NULL) return;
	root->keyRotate[0] += inc*(root->keyframef.x - root->keyframeb.x);
	root->keyRotate[1] += inc*(root->keyframef.y - root->keyframeb.y);
	root->keyRotate[2] += inc*(root->keyframef.z - root->keyframeb.z);
	// determine direction of rotation, and do the axes separately
	glPushMatrix();
	glLoadIdentity();
	// x axis
	if (root->keyframef.x - root->keyframeb.x > 0) {
		if (root->keyRotate[0] > root->keyframef.x) {
			float x = root->keyframef.x;
			root->keyframef.x = root->keyframeb.x;
			root->keyframeb.x = x;
		}
	}
	else if (root->keyframef.x - root->keyframeb.x < 0) {
		if (root->keyRotate[0] < root->keyframef.x) {
			float x = root->keyframef.x;
			root->keyframef.x = root->keyframeb.x;
			root->keyframeb.x = x;
		}
	}

	// y axis
	if (root->keyframef.y - root->keyframeb.y > 0) {
		if (root->keyRotate[1] > root->keyframef.y) {
			float y = root->keyframef.y;
			root->keyframef.y = root->keyframeb.y;
			root->keyframeb.y = y;
		}
	}
	else if (root->keyframef.y - root->keyframeb.y < 0) {
		if (root->keyRotate[1] < root->keyframef.y) {
			float y = root->keyframef.y;
			root->keyframef.y = root->keyframeb.y;
			root->keyframeb.y = y;
		}
	}

	// z axis
	if (root->keyframef.z - root->keyframeb.z > 0) {
		if (root->keyRotate[2] > root->keyframef.z) {
			float z = root->keyframef.z;
			root->keyframef.z = root->keyframeb.z;
			root->keyframeb.z = z;
		}
	}
	else if (root->keyframef.z - root->keyframeb.z < 0) {
		if (root->keyRotate[2] < root->keyframef.z) {
			float z = root->keyframef.z;
			root->keyframef.z = root->keyframeb.z;
			root->keyframeb.z = z;
		}
	}

	glRotatef(root->keyRotate[0],1,0,0);
	glRotatef(root->keyRotate[1],0,1,0);
	glRotatef(root->keyRotate[2],0,0,1);
	glGetFloatv(GL_MODELVIEW_MATRIX, root->animateMatrix);
	glPopMatrix();

	if (root->child != NULL) animateDelta(root->child,inc);

	if (root->sibling != NULL) animateDelta(root->sibling,inc);
}

void 
Object::traverse(Node* root)
{
	// traverse
	if (root == NULL) return;
	glPushMatrix();
	glMultMatrixf(root->animateMatrix);
	glMultMatrixf(root->localMatrix);
	glCallList(root->callstackid);
	if (root->child != NULL) traverse(root->child);
	glPopMatrix();
	if (root->sibling != NULL) traverse(root->sibling);
}

void 
Object::traverse_object() {
	if (!drawable) return;
	glPushMatrix();
	glPushName(objectName);
	glTranslatef(coordinates.x,coordinates.y,coordinates.z);
	traverse(head);
	glPopName();
	glPopMatrix();
}

void
Object::animateDelta_object() {
	if (!drawable) return;
	animateDelta(head,animateSpeed);
}

bool
Object:: hitByCoordsTraverse(Node* n, float& x, float& y, float& z) {
	if (n == NULL) return false;
	// do collision check
	bool retVal = false;

	if ( (abs(coordinates.x-x) < n->hitRadius) &&
		 (abs(coordinates.y-y) < n->hitRadius) &&
		 (abs(coordinates.z-z) < n->hitRadius) ) {
		return true;
	}

	if (n->child != NULL) {
		retVal = hitByCoordsTraverse(n->child,x,y,z);
	}
	
	if (retVal) return true;

	if (n->sibling != NULL) {
		retVal = hitByCoordsTraverse(n->sibling,x,y,z);	
	}
	if (retVal) return true;
	else return false;
}

bool
Object::hitByCoords(float x, float y, float z) {
	return hitByCoordsTraverse(head,x,y,z);
}

void
Object::kill() {
	shootTimer = 0;
	respawnTimer = 0;
	drawable = false;
}

void
Object::respawn() {
	shootTimer = 0;
	respawnTimer = 0;
	hitpoints = worth;
	drawable = true;
}