#ifndef _OBJECT_
#define _OBJECT_

#include "glut.h"
#include <iostream.h>
#include <fstream.h>

static int nameCount = 1;	

typedef struct {
	float x;
	float y;
	float z;
} triple;

class Node {
public:
	GLfloat localMatrix[16];
	GLfloat animateMatrix[16];
	Node* sibling;
	Node* child;
	GLfloat keyRotate[3];
	int callstackid;
	triple keyframef;
	triple keyframeb;
	float hitRadius;
	Node();
	void loadMesh(char* fname);
};

class Object {
private:
	float localTransform[16];
	float worldTransform[16];
	int callstackNumber;
	bool hittable;
	void destroy(Node* n);
	void animateDelta(Node* root, float inc);
	void traverse(Node* root);
	void readObjectConfig(Node *parent, ifstream &input);
	void copyNodeHelper(Node* newNode, Node* oldNode);

protected:
	triple coordinates;
public:
	Node* head;
	int objectName;
	bool animated;
	bool movable;
	float animateSpeed;
	int hitpoints;
	bool exploding;
	bool drawable;
	int worth;
	int shootDelay;
	int shootTimer;
	int respawnDelay;
	int respawnTimer;

	Object();
	virtual ~Object();
	Object(Object& o);
	void setLocalMatrix(float* m);
	void setWorldMatrix(float* m);
	void setHittable(bool hit);
	void setCallStackID(int id);
	void setCoordinates(float x, float y, float z);
	void setCoordinates(triple c);
	float* getLocalMatrix();
	float* getWorldMatrix();
	bool isHittable();
	int getObjectName();
	void setObjectName(int name);
	void copy(Object* o);
	
	void respawn(); 
	void kill();
	int getCallStackID();
	triple getCoordinates();
	bool hitByCoordsTraverse(Node* n, float& x, float& y, float& z);
	bool hitByCoords(float x, float y, float z);
	void addHead(Node* n);
	void animateDelta_object();
	void traverse_object();
	void readFile(char *fname); 

};

#endif