#include <vector>
#include <iostream.h>
#include "ObjectList.h"

ObjectList::ObjectList() {
	player = NULL;
	missile1Template = NULL;
	missile2Template = NULL;
	score = 0;
}

ObjectList::~ObjectList() {
	int i = 0;
	for (i = 0; i < objects.size(); i++) {
		Object* ref = objects[i];
		delete ref;
	}
	objects.clear();
	for (i = 0; i < missiles.size(); i++) {
		Mobile* ref = missiles[i];
		delete ref;
	}
	missiles.clear();
	delete missile1Template;
	delete missile2Template;
	delete player;
}

void 
ObjectList::playerCoords(float& x, float& y, float&z) {
	triple coords = player->getCoordinates();
	x = coords.x;
	y = coords.y;
	z = coords.z;
}

int 
ObjectList::playerHitpoints() {
	return player->hitpoints;
}

int 
ObjectList::playerWorth() {
	return player->worth;
}

void 
ObjectList::loadHelper(ifstream& input) {
	char token[512];
	Object* o = NULL;
	bool player = false;
	bool missile1 = false;
	bool missile2 = false;
	while ( !input.eof() ) {
		input >> token;
		if ( token[0] == '#' ) {
			input.getline( token, 512, '\n' );
		} else if ( strcmp( token, "mobile" ) == 0 ) {
			input >> token; 
			o = new Mobile();
		} else if ( strcmp( token, "player" ) == 0 ) {
			input >> token; 
			player = true;
			o = new Mobile();
		} else if ( strcmp( token, "missile1" ) == 0 ) {
			input >> token;
			missile1 = true;
			o = new Mobile();
		} else if ( strcmp( token, "missile2" ) == 0 ) {
			input >> token;
			missile2 = true;
			o = new Mobile();
		} else if ( strcmp( token, "object" ) == 0 ) {
			input >> token; 
			o = new Object();
		} else if ( strcmp( token, "file" ) == 0 ) {
			char fname[512];
			input >> fname; 
			o->readFile(fname);
		} else if ( strcmp( token, "coordinates" ) == 0 ) {
			float x, y, z;
			input >> x >> y >> z; 
			o->setCoordinates(x,y,z);
		} else if ( strcmp( token, "animatespeed" ) == 0 ) {
			float x;
			input >> x; 
			o->animateSpeed = x;
		} else if ( strcmp( token, "setpath" ) == 0 ) {
			float x, y, z;
			input >> x >> y >> z; 
			((Mobile*)o)->addToPath(x,y,z);
		} else if ( strcmp( token, "movespeed" ) == 0 ) {
			float x;
			input >> x; 
			((Mobile*)o)->moveSpeed = x;
		} else if ( strcmp( token, "hitpoints" ) == 0 ) {
			float x;
			input >> x; 
			o->hitpoints = x;
		} else if ( strcmp( token, "worth" ) == 0 ) {
			float x;
			input >> x; 
			o->worth = x;
		} else if ( strcmp( token, "yield" ) == 0 ) {
			float x;
			input >> x; 
			((Mobile*)o)->yield = x;
		} else if ( strcmp( token, "shootdelay" ) == 0 ) {
			float x;
			input >> x; 
			o->shootDelay = x;
		} else if ( strcmp( token, "respawndelay" ) == 0 ) {
			float x;
			input >> x; 
			o->respawnDelay = x;
		} else if ( strcmp( token, "}" ) == 0 ) {
			if (player) {
				this->player = (Mobile*)o;
				this->player->head = new Node();
				player = false;
			}
			else if (missile1) {
				this->missile1Template = (Mobile*) o;
				missile1 = false;
			}
			else if (missile2) {
				this->missile2Template = (Mobile*) o;
				missile2 = false;
			}
			else {
				objects.push_back(o);
			}
			o = NULL;
		}
	}

}


void 
ObjectList::loadFile(char* fname) {
	ifstream input(fname);
	if ( !input.is_open() ) {
		printf("couldn't open");
		exit(1);
	}
	loadHelper( input );
	input.close();
}

void
ObjectList::move() {
	player->moveCoordinatesToVector();
	int i = 0;
	for (i = 0; i < objects.size(); i++) {
		Object* ref = objects[i];
		if (ref->animated) {
			ref->animateDelta_object();
		}
		if (ref->movable) {
			((Mobile*)ref)->moveCoordinatesToVector();
		}
	}
	for (i = 0; i < missiles.size(); i++ ) {
		Object* ref = missiles[i];
		if (ref->movable) {
			((Mobile*)ref)->moveCoordinatesToVector();
			int kill = 0;
			if ( ((Mobile*)ref)->checkHit(kill) ) {
				missiles.erase(missiles.begin() + i);
				if (kill) {
					score += kill;
				}
			}
		}
		if (ref->animated) {
			ref->animateDelta_object();
		}
	}
}

void 
ObjectList::draw() {
	int i = 0;
	for (i = 0; i < missiles.size(); i++) {
		Object* ref = missiles[i];

		ref->traverse_object();
	}	
	for (i = 0; i < objects.size(); i++) {
		Object* ref = objects[i];
		ref->traverse_object();
	}
}

void 
ObjectList::addMissile(float x, float y, float z, int target) {
	Mobile* m = new Mobile();
	int i = 0;

	// if target == 0, then the player is the target
	if (target == 0) {
		m->target = player;
	}
	// do a search and find the target object
	else {
		bool found = false;
		for (i = 0; i < objects.size(); i++) {
			if ( (objects[i]->objectName) == target) {
				m->target = objects[i];
				found = true;
				break;
			}
		}
		// not found in object, just return
		if (!found) return;
	}

	if (target != 0) {
		m->copy(missile1Template);
		// copy missile stuff
		m->moveSpeed = missile1Template->moveSpeed;
		m->setCoordinates(x,y,z);
		m->missile = true;
		m->movable = true;
		m->yield = missile1Template->yield;
		m->worth = missile1Template->worth;
		// missile is owned by player if target != 0
		m->owner = 0;
	}
	else {
		m->copy(missile2Template);
		// copy missile stuff
		m->moveSpeed = missile2Template->moveSpeed;
		m->setCoordinates(x,y,z);
		m->missile = true;
		m->movable = true;
		m->yield = missile2Template->yield;
		m->worth = missile2Template->worth;
		// missile is owned by player if target != 0
		m->owner = 1;
	}

	missiles.push_back(m);
}

bool
ObjectList::targetCoords(int& name,float& x,float& y,float& z) {
	for (int i = 0; i < objects.size(); i++) {
		if ( (objects[i]->objectName) == name) {
			triple c = objects[i]->getCoordinates();
			x = c.x;
			y = c.y;
			z = c.z;
			return true;
		}
	}
	// not found in object, just return
	return false;
}

void 
ObjectList::hitEntity(int name) {
// do a search through all objects
	int i = 0;
	for (i = 0; i < objects.size(); i++) {
		if ( (objects[i]->objectName) == name) {
			objects[i]->hitpoints -= 5;
			if ( (objects[i]->hitpoints <= 0) && (objects[i]->drawable) ) {
				objects[i]->kill();
				score += objects[i]->worth;
			}
			return;
		}
	}
// do a search for all missiles
	for (i = 0; i < missiles.size(); i++) {
		if ( ((missiles[i]->objectName) == name) && (missiles[i]->owner != 0) ) {
			missiles[i]->hitpoints -= 5;
			if (missiles[i]->hitpoints <= 0) {
				missiles[i]->kill();
				missiles.erase(missiles.begin() + i);
				score += missiles[i]->worth;
			}
			return;
		}
	}
}

void 
ObjectList::updateTimers() {
// do a search through all objects
	int i = 0;
	for (i = 0; i < objects.size(); i++) {
		if ( (objects[i]->shootDelay > 0) && (objects[i]->drawable) ) {
			objects[i]->shootTimer++;
			if (objects[i]->shootDelay == objects[i]->shootTimer) {
				triple objCoords = objects[i]->getCoordinates();
				this->addMissile(objCoords.x,objCoords.y,objCoords.z,0);
				objects[i]->shootTimer = 0;
			}
		}
		if ( (objects[i]->respawnDelay > 0)  && (!objects[i]->drawable) ) {
			objects[i]->respawnTimer++;
			if (objects[i]->respawnDelay == objects[i]->respawnTimer) {
				objects[i]->respawn();
				objects[i]->respawnTimer = 0;
			}
		}		
	}	
}

long
ObjectList::getScore() {
	return score;
}