#include "ObjectNode.h"
#include "glm.h"

Node::Node() {
	for (int i = 0; i < 16; i++) {
		localMatrix[i] = 0;
		animateMatrix[i] = 0;
	}
	localMatrix[0] = localMatrix[5] = localMatrix[10] = localMatrix[15] = 1;
	animateMatrix[0] = animateMatrix[5] = animateMatrix[10] = animateMatrix[15] = 1;
	sibling = NULL;
	child = NULL;
	keyRotate[0] = keyRotate[1] = keyRotate[2] = 0;	
	callstackid = 0;
}

void
Node::loadMesh(char* fname) {
  GLMmodel* enemy = glmReadOBJ(fname);
  float scale = glmUnitize(enemy);
  glmFacetNormals(enemy);
  glmVertexNormals(enemy, 90.0);
  glmScale(enemy,3.0);
  callstackid = glmList(enemy,GLM_SMOOTH | GLM_MATERIAL);
  glmDelete(enemy);
}

ObjectHierarchy::ObjectHierarchy() {

}

ObjectHierarchy::~ObjectHierarchy() {
	destroy(head);
}

void 
ObjectHierarchy::destroy(Node* n) {
	// traverse
	if (n == NULL) return;
	if (n->child != NULL) destroy(n->child);
	delete n->child;
	if (n->sibling != NULL) traverse(n->sibling);
	delete n->sibling;
}

void
ObjectHierarchy::addHead(Node* n) {
	head = n;
}

void 
ObjectHierarchy::animateDelta(Node* root,float inc) {
	// traverse
	// traverse
	if (root == NULL) return;
	glPushMatrix();
	glLoadIdentity();
	glMultMatrixf(root->animateMatrix);
	glRotatef(1,0,0,inc*root->keyRotate[0]);
	glRotatef(0,1,0,inc*root->keyRotate[1]);
	glRotatef(0,0,1,inc*root->keyRotate[2]);	
	glGetFloatv(GL_MODELVIEW_MATRIX, root->animateMatrix);
	if (root->child != NULL) traverse(root->child);
	glPopMatrix();
	if (root->sibling != NULL) traverse(root->sibling);
}

void 
ObjectHierarchy::traverse(Node* root)
{
	// traverse
	if (root == NULL) return;
	glPushMatrix();
	glMultMatrixf(root->localMatrix);
	glMultMatrixf(root->animateMatrix);
	glCallList(root->callstackid);
	if (root->child != NULL) traverse(root->child);
	glPopMatrix();
	if (root->sibling != NULL) traverse(root->sibling);
}

void 
ObjectHierarchy::traverse_object() {
	traverse(head);
}

void
ObjectHierarchy::animateDelta_object(float inc) {
	animateDelta(head,inc);
}
