#include "mouse.h"
#include <iostream.h>

Mouse::Mouse() {

}

Mouse::~Mouse() {
	g_pMouse->Unacquire();
    
	// Release any DirectInput objects.
	g_pMouse->Release();
	g_pMouse = NULL;
	g_lpDI->Release();
	g_lpDI = NULL;
}

void 
Mouse::getMouseMovement(int& x, int& y, int& z, int& b1,int& b2) {
    HRESULT       hr;
    DIMOUSESTATE2 dims2;      // DirectInput mouse state structure

    if( NULL == g_pMouse ) 
        return;
    
    // Get the input's device state, and put the state in dims
    ZeroMemory( &dims2, sizeof(dims2) );
    hr = g_pMouse->GetDeviceState( sizeof(DIMOUSESTATE2), &dims2 );
    if( FAILED(hr) ) 
    {
        // DirectInput may be telling us that the input stream has been
        // interrupted.  We aren't tracking any state between polls, so
        // we don't have any special reset that needs to be done.
        // We just re-acquire and try again.
        
        // If input is lost then acquire and keep trying 
        hr = g_pMouse->Acquire();
        while( hr == DIERR_INPUTLOST ) 
            hr = g_pMouse->Acquire();

        return; 
    }
    
    // The dims structure now has the state of the mouse, so 
    // display mouse coordinates (x, y, z) and buttons.
	x = dims2.lX;
	y = dims2.lY;
	z = dims2.lZ;
	if ( (dims2.rgbButtons[0] & 0x80) ) {
		b1 = 1;
	}
	if ( (dims2.rgbButtons[1] & 0x80) ) {
		b2 = 1;
	}
/*
    _stprintf( strNewText, TEXT("(X=% 3.3d, Y=% 3.3d, Z=% 3.3d) B0=%c B1=%c B2=%c B3=%c B4=%c B5=%c B6=%c B7=%c"),
                         dims2.lX, dims2.lY, dims2.lZ,
                        (dims2.rgbButtons[0] & 0x80) ? '1' : '0',
                        (dims2.rgbButtons[1] & 0x80) ? '1' : '0',
                        (dims2.rgbButtons[2] & 0x80) ? '1' : '0',
                        (dims2.rgbButtons[3] & 0x80) ? '1' : '0',
                        (dims2.rgbButtons[4] & 0x80) ? '1' : '0',
                        (dims2.rgbButtons[5] & 0x80) ? '1' : '0',
                        (dims2.rgbButtons[6] & 0x80) ? '1' : '0',
                        (dims2.rgbButtons[7] & 0x80) ? '1' : '0');

*/
    return;
}

void 
Mouse::init(HWND hWnd) {
	this->hWnd = hWnd;
	hr = DirectInput8Create(GetModuleHandle(NULL), DIRECTINPUT_VERSION, 
        IID_IDirectInput8, (void**)&g_lpDI, NULL); 
	if FAILED(hr) 
	{ 
		exit(1);
		// DirectInput not available; take appropriate action 
	} 
	hr = g_lpDI->CreateDevice(GUID_SysMouse, &g_pMouse, NULL);
 
	if (FAILED(hr)) {
		exit(1);
	}

	hr = g_pMouse->SetDataFormat(&c_dfDIMouse2);
 
	if (FAILED(hr)) {
		exit(1);
	}

	hr = g_pMouse->SetCooperativeLevel(hWnd,
               DISCL_EXCLUSIVE | DISCL_FOREGROUND);
 
	if (FAILED(hr)) {
		exit(1);
	}

    g_pMouse->Acquire();
}